import abc
import itertools
import operator
import math
from wntr.utils.ordered_set import OrderedSet
import enum
from six import with_metaclass

if not hasattr(math, 'inf'):
    math.inf = float('inf')

native_numeric_types = {float, int}
native_integer_types = {int, bool}
native_boolean_types = {int, bool, str}


class OperationEnum(enum.IntEnum):
    add = -1
    sub = -2
    mul = -3
    div = -4
    pow = -5
    abs = -6
    sign = -7
    if_else = -8
    inequality = -9
    exp = -10
    log = -11
    negation = -12
    sin = -13
    cos = -14
    tan = -15
    asin = -16
    acos = -17
    atan = -18


class Node(with_metaclass(abc.ABCMeta, object)):

    __slots__ = ()

    @abc.abstractmethod
    def is_leaf(self):
        pass


class ExpressionBase(with_metaclass(abc.ABCMeta, Node)):
    """
    A base class for expressions (including variables and params).
    """
    __slots__ = ()

    def is_relational(self):
        return False

    @abc.abstractmethod
    def operators(self):
        pass

    @abc.abstractmethod
    def last_node(self):
        pass

    @abc.abstractmethod
    def evaluate(self):
        """
        Evaluate the expression numerically.

        Returns
        -------
        val: float
            The floating point value of the expression.
        """
        pass

    @abc.abstractmethod
    def _binary_operation_helper(self, other, cls):
        pass

    @abc.abstractmethod
    def _unary_operation_helper(self, cls):
        pass

    def __add__(self, other):
        if other == 0:
            return self
        return self._binary_operation_helper(other, AddOperator)

    def __sub__(self, other):
        if other == 0:
            return self
        return self._binary_operation_helper(other, SubtractOperator)

    def __mul__(self, other):
        if other == 0:
            return 0
        elif other == 1:
            return self
        return self._binary_operation_helper(other, MultiplyOperator)

    def __truediv__(self, other):
        if other == 0:
            raise ValueError('Divide by 0')
        elif other == 1:
            return self
        return self._binary_operation_helper(other, DivideOperator)

    def __div__(self, other):
        if other == 0:
            raise ValueError('Divide by 0')
        elif other == 1:
            return self
        return self._binary_operation_helper(other, DivideOperator)

    def __pow__(self, other):
        if other == 0:
            return 1
        elif other == 1:
            return self
        return self._binary_operation_helper(other, PowerOperator)

    def __radd__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return self
        return Float(other) + self

    def __rsub__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return -self
        return Float(other) - self

    def __rmul__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return 0
        elif other == 1:
            return self
        return Float(other) * self

    def __rtruediv__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return 0
        return Float(other) / self

    def __rdiv__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return 0
        return Float(other) / self

    def __rpow__(self, other):
        assert type(other) in native_numeric_types
        if other == 0:
            return 0
        elif other == 1:
            return 1
        return Float(other) ** self

    def __neg__(self):
        return self._unary_operation_helper(NegationOperator)

    @abc.abstractmethod
    def get_vars(self):
        pass

    @abc.abstractmethod
    def get_params(self):
        pass

    @abc.abstractmethod
    def get_floats(self):
        pass

    @abc.abstractmethod
    def get_leaves(self):
        pass

    @abc.abstractmethod
    def is_parameter_type(self):
        pass

    @abc.abstractmethod
    def is_variable_type(self):
        pass

    @abc.abstractmethod
    def is_float_type(self):
        pass

    @abc.abstractmethod
    def is_expression_type(self):
        pass

    @abc.abstractmethod
    def reverse_ad(self):
        pass

    @abc.abstractmethod
    def reverse_sd(self):
        pass

    @abc.abstractmethod
    def get_rpn(self, leaf_ndx_map):
        pass

    def __repr__(self):
        return str(self)


class Leaf(with_metaclass(abc.ABCMeta, ExpressionBase)):

    __slots__ = ('_value', '_c_obj')

    def is_leaf(self):
        return True

    def _binary_operation_helper(self, other, cls):
        if type(other) in native_numeric_types:
            other = Float(other)
        new_operator = cls(self, other.last_node())
        if other.is_leaf():
            expr = expression()
        else:
            expr = expression(other)
        expr.append_operator(new_operator)
        return expr

    def _unary_operation_helper(self, cls):
        new_operator = cls(self)
        expr = expression()
        expr.append_operator(new_operator)
        return expr

    def last_node(self):
        return self

    def operators(self):
        return list()

    @property
    def value(self):
        if self._c_obj is not None:
            return self._c_obj.value
        return self._value

    @value.setter
    def value(self, val):
        self._value = val
        if self._c_obj is not None:
            self._c_obj.value = val

    def evaluate(self):
        return self._value

    @abc.abstractmethod
    def _str(self):
        pass

    def __str__(self):
        return self._str()

    def reverse_ad(self):
        return {self: 1}

    def reverse_sd(self):
        return {self: 1}

    def get_rpn(self, leaf_ndx_map):
        return [leaf_ndx_map[self]]


class Float(Leaf):

    __slots__ = ()

    def __init__(self, val):
        self._value = val
        self._c_obj = None

    def is_parameter_type(self):
        return False

    def is_variable_type(self):
        return False

    def is_float_type(self):
        return True

    def is_expression_type(self):
        return False

    def _str(self):
        return str(self.value)

    def get_vars(self):
        return OrderedSet()

    def get_params(self):
        return OrderedSet()

    def get_floats(self):
        return OrderedSet([self])

    def get_leaves(self):
        return OrderedSet([self])

    def _binary_operation_helper(self, other, cls):
        if type(other) in native_numeric_types:
            return cls.operation(self.value, other)
        elif other.is_float_type():
            return cls.operation(self.value, other.value)
        new_operator = cls(self, other.last_node())
        if other.is_leaf():
            expr = expression()
        else:
            expr = expression(other)
        expr.append_operator(new_operator)
        return expr

    def _unary_operation_helper(self, cls):
        return cls.operation(self.value)


class Var(Leaf):
    """
    Variables

    Parameters
    ----------
    val: float
        value of the variable
    """

    __slots__ = ('_name',)

    def __init__(self, val=0):
        self._value = val
        self._name = None
        self._c_obj = None

    def is_parameter_type(self):
        return False

    def is_variable_type(self):
        return True

    def is_float_type(self):
        return False

    def is_expression_type(self):
        return False

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    def _str(self):
        return str(self.name)

    def get_vars(self):
        return OrderedSet([self])

    def get_params(self):
        return OrderedSet()

    def get_floats(self):
        return OrderedSet()

    def get_leaves(self):
        return OrderedSet([self])

    @property
    def index(self):
        if self._c_obj is None:
            return None
        else:
            return self._c_obj.index


class Param(Leaf):

    __slots__ = ('_name',)

    def __init__(self, val=0):
        self._value = val
        self._name = None
        self._c_obj = None

    def is_parameter_type(self):
        return True

    def is_variable_type(self):
        return False

    def is_float_type(self):
        return False

    def is_expression_type(self):
        return False

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    def _str(self):
        return str(self.name)

    def get_vars(self):
        return OrderedSet()

    def get_params(self):
        return OrderedSet([self])

    def get_floats(self):
        return OrderedSet()

    def get_leaves(self):
        return OrderedSet([self])


class expression(ExpressionBase):

    __slots__ = ('_operators', '_n_opers', '_vars', '_params', '_floats')

    def __init__(self, expr=None):
        """

        Parameters
        ----------
        expr: expression
        """
        if expr is not None:
            if expr._operators[-1] is not expr.last_node():
                self._operators = expr.list_of_operators()
            else:
                self._operators = expr._operators
        else:
            self._operators = []
        self._n_opers = len(self._operators)
        self._vars = None
        self._params = None
        self._floats = None

    def append_operator(self, oper):
        self._operators.append(oper)
        self._n_opers += 1

    def last_node(self):
        """
        Returns
        -------
        last_node: Operator
        """
        return self._operators[self._n_opers - 1]

    def list_of_operators(self):
        return self._operators[:self._n_opers]

    def is_leaf(self):
        return False

    def operators(self):
        return itertools.islice(self._operators, 0, self._n_opers)

    def _binary_operation_helper(self, other, cls):
        if type(other) in native_numeric_types:
            other = Float(other)
        new_operator = cls(self.last_node(), other.last_node())
        expr = expression(self)
        for oper in other.operators():
            expr.append_operator(oper)
        expr.append_operator(new_operator)
        return expr

    def _unary_operation_helper(self, cls):
        new_operator = cls(self.last_node())
        expr = expression(self)
        expr.append_operator(new_operator)
        return expr

    def evaluate(self):
        val_dict = dict()
        for oper in self.operators():
            oper.evaluate(val_dict)
        return val_dict[self.last_node()]

    def get_vars(self):
        if self._vars is None:
            self._collect_leaves()
        for i in self._vars:
            yield i

    def get_params(self):
        if self._params is None:
            self._collect_leaves()
        for i in self._params:
            yield i

    def get_floats(self):
        if self._floats is None:
            self._collect_leaves()
        for i in self._floats:
            yield i

    def _collect_leaves(self):
        self._vars = OrderedSet()
        self._params = OrderedSet()
        self._floats = OrderedSet()
        for oper in self.operators():
            for operand in oper.operands():
                if operand.is_leaf():
                    if operand.is_variable_type():
                        self._vars.add(operand)
                    elif operand.is_parameter_type():
                        self._params.add(operand)
                    elif operand.is_float_type():
                        self._floats.add(operand)
                    elif operand.is_expression_type():
                        self._vars.update(operand.get_vars())
                        self._params.update(operand.get_params())
                        self._floats.update(operand.get_floats())
                    else:
                        raise ValueError('operand type not recognized: ' + str(operand))

    def get_leaves(self):
        if self._vars is None:
            self._collect_leaves()
        for i in self._vars:
            yield i
        for i in self._params:
            yield i
        for i in self._floats:
            yield i

    def _str(self):
        return str(self)

    def __str__(self):
        val_dict = dict()
        for oper in self.operators():
            oper._str(val_dict)
        return val_dict[self.last_node()]

    def is_variable_type(self):
        return False

    def is_parameter_type(self):
        return False

    def is_float_type(self):
        return False

    def is_expression_type(self):
        return True

    def reverse_ad(self):
        val_dict = dict()
        der_dict = dict()
        for oper in self.operators():
            oper.diff_up(val_dict, der_dict)
        der_dict[self.last_node()] = 1
        for oper in reversed(self.list_of_operators()):
            oper.diff_down(val_dict, der_dict)
        return der_dict

    def reverse_sd(self):
        val_dict = dict()
        der_dict = dict()
        for oper in self.operators():
            oper.diff_up_symbolic(val_dict, der_dict)
        der_dict[self.last_node()] = 1
        for oper in reversed(self.list_of_operators()):
            oper.diff_down(val_dict, der_dict)
        return der_dict

    def is_relational(self):
        if type(self.last_node()) in {InequalityOperator}:
            return True
        return False

    def get_rpn(self, leaf_ndx_map):
        rpn_map = dict()
        for oper in self.operators():
            oper.get_rpn(rpn_map, leaf_ndx_map)
        return rpn_map[self.last_node()]


class Operator(with_metaclass(abc.ABCMeta, Node)):

    __slots__ = ()

    def is_leaf(self):
        return False

    @abc.abstractmethod
    def evaluate(self, val_dict):
        pass

    @abc.abstractmethod
    def operands(self):
        pass

    @abc.abstractmethod
    def diff_up(self, val_dict, der_dict):
        pass

    @abc.abstractmethod
    def diff_down(self, val_dict, der_dict):
        pass

    @abc.abstractmethod
    def diff_up_symbolic(self, val_dict, der_dict):
        pass

    @abc.abstractmethod
    def get_rpn(self, rpn_map, leaf_ndx_map):
        pass


class BinaryOperator(Operator):

    __slots__ = ('_operand1', '_operand2')

    operation = None
    str_repn = None
    operation_enum = None

    def __init__(self, operand1, operand2):
        self._operand1 = operand1
        self._operand2 = operand2

    def evaluate(self, val_dict):
        if self._operand1.is_leaf():
            val1 = self._operand1.value
        else:
            val1 = val_dict[self._operand1]
        if self._operand2.is_leaf():
            val2 = self._operand2.value
        else:
            val2 = val_dict[self._operand2]
        val_dict[self] = self.operation(val1, val2)

    def _str(self, val_dict):
        if self._operand1.is_leaf():
            val1 = self._operand1._str()
        else:
            val1 = val_dict[self._operand1]
        if self._operand2.is_leaf():
            val2 = self._operand2._str()
        else:
            val2 = val_dict[self._operand2]
        val_dict[self] = '(' + val1 + self.str_repn + val2 + ')'

    def operands(self):
        yield self._operand1
        yield self._operand2

    def diff_up(self, val_dict, der_dict):
        if self._operand1.is_leaf():
            val1 = self._operand1.value
            val_dict[self._operand1] = val1
            if self._operand1 not in der_dict:
                der_dict[self._operand1] = 0
        else:
            val1 = val_dict[self._operand1]
            der_dict[self._operand1] = 0
        if self._operand2.is_leaf():
            val2 = self._operand2.value
            val_dict[self._operand2] = val2
            if self._operand2 not in der_dict:
                der_dict[self._operand2] = 0
        else:
            val2 = val_dict[self._operand2]
            der_dict[self._operand2] = 0
        val_dict[self] = self.operation(val1, val2)

    def diff_up_symbolic(self, val_dict, der_dict):
        if self._operand1.is_leaf():
            val1 = self._operand1
            val_dict[self._operand1] = val1
            if self._operand1 not in der_dict:
                der_dict[self._operand1] = 0
        else:
            val1 = val_dict[self._operand1]
            der_dict[self._operand1] = 0
        if self._operand2.is_leaf():
            val2 = self._operand2
            val_dict[self._operand2] = val2
            if self._operand2 not in der_dict:
                der_dict[self._operand2] = 0
        else:
            val2 = val_dict[self._operand2]
            der_dict[self._operand2] = 0
        val_dict[self] = self.operation(val1, val2)

    def get_rpn(self, rpn_map, leaf_ndx_map):
        if self._operand2.is_leaf():
            if self._operand1.is_leaf():
                rpn_map[self] = [leaf_ndx_map[self._operand1], leaf_ndx_map[self._operand2], self.operation_enum]
            else:
                rpn_map[self] = _rpn = rpn_map[self._operand1]
                _rpn.append(leaf_ndx_map[self._operand2])
                _rpn.append(self.operation_enum)
        elif self._operand1.is_leaf():
            rpn_map[self] = _rpn = rpn_map[self._operand2]
            _rpn.insert(0, leaf_ndx_map[self._operand1])
            _rpn.append(self.operation_enum)
        else:
            rpn_map[self] = _rpn = rpn_map[self._operand1]
            _rpn.extend(rpn_map[self._operand2])
            _rpn.append(self.operation_enum)


class AddOperator(BinaryOperator):

    __slots__ = ()

    operation = operator.add
    str_repn = '+'
    operation_enum = OperationEnum.add.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand1] += der
        der_dict[self._operand2] += der


class SubtractOperator(BinaryOperator):
    __slots__ = ()

    operation = operator.sub
    str_repn = '-'
    operation_enum = OperationEnum.sub.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand1] += der
        der_dict[self._operand2] -= der


class MultiplyOperator(BinaryOperator):
    __slots__ = ()

    operation = operator.mul
    str_repn = '*'
    operation_enum = OperationEnum.mul.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand1] += der * val_dict[self._operand2]
        der_dict[self._operand2] += der * val_dict[self._operand1]


class DivideOperator(BinaryOperator):
    __slots__ = ()

    operation = operator.truediv
    str_repn = '/'
    operation_enum = OperationEnum.div.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand1] += der / val_dict[self._operand2]
        der_dict[self._operand2] -= der * val_dict[self._operand1] / val_dict[self._operand2]**2


class PowerOperator(BinaryOperator):
    __slots__ = ()

    operation = operator.pow
    str_repn = '**'
    operation_enum = OperationEnum.pow.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        val1 = val_dict[self._operand1]
        val2 = val_dict[self._operand2]
        der_dict[self._operand1] += der * val2 * val1**(val2 - 1)
        if not self._operand2.is_leaf() or self._operand2.is_variable_type():
            der_dict[self._operand2] += der * val1**val2 * log(val1)


class UnaryOperator(Operator):
    __slots__ = ('_operand',)

    str_repn = None
    operation_enum = None

    def __init__(self, operand):
        """
        Parameters
        ----------
        operand: Node
        """
        self._operand = operand

    def evaluate(self, val_dict):
        if self._operand.is_leaf():
            val = self._operand.value
        else:
            val = val_dict[self._operand]
        val_dict[self] = self.operation(val)

    def _str(self, val_dict):
        if self._operand.is_leaf():
            val = self._operand._str()
        else:
            val = val_dict[self._operand]
        val_dict[self] = '(' + self.str_repn + '(' + val + ')' + ')'

    def operands(self):
        yield self._operand

    def diff_up(self, val_dict, der_dict):
        if self._operand.is_leaf():
            val = self._operand.value
            val_dict[self._operand] = val
            if self._operand not in der_dict:
                der_dict[self._operand] = 0
        else:
            val = val_dict[self._operand]
            der_dict[self._operand] = 0
        val_dict[self] = self.operation(val)

    def diff_up_symbolic(self, val_dict, der_dict):
        if self._operand.is_leaf():
            val = self._operand
            val_dict[self._operand] = val
            if self._operand not in der_dict:
                der_dict[self._operand] = 0
        else:
            val = val_dict[self._operand]
            der_dict[self._operand] = 0
        val_dict[self] = self.operation(val)

    def get_rpn(self, rpn_map, leaf_ndx_map):
        if self._operand.is_leaf():
            rpn_map[self] = [leaf_ndx_map[self._operand], self.operation_enum]
        else:
            rpn_map[self] = _rpn = rpn_map[self._operand]
            _rpn.append(self.operation_enum)

    @staticmethod
    def operation(val):
        raise NotImplementedError('Subclasses should implement this.')


class NegationOperator(UnaryOperator):
    __slots__ = ()

    str_repn = '-'
    operation_enum = OperationEnum.negation.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] -= der

    @staticmethod
    def operation(val):
        return -val


def exp(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.exp(val)
    return val._unary_operation_helper(ExpOperator)


def log(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.log(val)
    return val._unary_operation_helper(LogOperator)


def sin(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.sin(val)
    return val._unary_operation_helper(SinOperator)


def cos(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.cos(val)
    return val._unary_operation_helper(CosOperator)


def tan(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.tan(val)
    return val._unary_operation_helper(TanOperator)


def asin(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.asin(val)
    return val._unary_operation_helper(AsinOperator)


def acos(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.acos(val)
    return val._unary_operation_helper(AcosOperator)


def atan(val):
    """

    Parameters
    ----------
    val: ExpressionBase

    Returns
    -------
    expr: expression
    """
    if type(val) in native_numeric_types:
        return math.atan(val)
    return val._unary_operation_helper(AtanOperator)


def if_else(if_statement, then_statement, else_statement):
    """

    Parameters
    ----------
    if_statement: ExpressionBase
    then_statement: ExpressionBase
    else_statement: ExpressionBase

    Returns
    -------
    expr: ExpressionBase
    """
    if type(if_statement) in native_numeric_types or type(if_statement) in native_boolean_types:
        if if_statement:
            return then_statement
        else:
            return else_statement
    if type(then_statement) in native_numeric_types:
        then_statement = Float(then_statement)
    if type(else_statement) in native_numeric_types:
        else_statement = Float(else_statement)
    assert if_statement.is_relational()
    expr = expression(if_statement)
    new_operator = IfElseOperator(if_statement.last_node(), then_statement.last_node(), else_statement.last_node())
    for oper in then_statement.operators():
        expr.append_operator(oper)
    for oper in else_statement.operators():
        expr.append_operator(oper)
    expr.append_operator(new_operator)
    return expr


def inequality(body, lb=None, ub=None):
    """

    Parameters
    ----------
    body: ExpressionBase or float
    lb: float
    ub: float

    Returns
    -------
    expr: ExpressionBase
    """
    if lb is not None and ub is not None:
        if type(lb) not in native_numeric_types or type(ub) not in native_numeric_types:
            raise ValueError('inequality can only accept both lb and ub arguments if both are floats or ints.')

    if lb is None:
        lb = -math.inf
    if ub is None:
        ub = math.inf

    if type(lb) in native_numeric_types:
        lb = Float(lb)
    else:
        body -= lb
        lb = Float(0)

    if type(ub) in native_numeric_types:
        ub = Float(ub)
    else:
        body -= ub
        ub = Float(0)

    if type(body) in native_numeric_types:
        return lb.value <= body <= ub.value

    if body.is_leaf():
        expr = expression()
    else:
        expr = expression(body)
    new_operator = InequalityOperator(body.last_node(), lb=lb, ub=ub)
    expr.append_operator(new_operator)
    return expr


def sign(val):
    if type(val) in native_numeric_types:
        if val >= 0:
            return 1
        else:
            return -1
    return val._unary_operation_helper(SignOperator)


def abs(val):
    if type(val) in native_numeric_types:
        return math.fabs(val)
    return val._unary_operation_helper(AbsOperator)


class IfElseOperator(Operator):
    __slots__ = ('_if_arg', '_then_arg', '_else_arg')

    def __init__(self, if_arg, then_arg, else_arg):
        self._if_arg = if_arg
        self._then_arg = then_arg
        self._else_arg = else_arg

    def evaluate(self, val_dict):
        if_val = val_dict[self._if_arg]
        if if_val:
            if self._then_arg.is_leaf():
                res = self._then_arg.value
            else:
                res = val_dict[self._then_arg]
        else:
            if self._else_arg.is_leaf():
                res = self._else_arg.value
            else:
                res = val_dict[self._else_arg]
        val_dict[self] = res

    def operands(self):
        yield self._if_arg
        yield self._then_arg
        yield self._else_arg

    def _str(self, val_dict):
        if_val = val_dict[self._if_arg]
        if self._then_arg.is_leaf():
            then_str = self._then_arg._str()
        else:
            then_str = val_dict[self._then_arg]
        if self._else_arg.is_leaf():
            else_str = self._else_arg._str()
        else:
            else_str = val_dict[self._else_arg]
        s = '\n (\n'
        s += '  if ' + if_val + ':\n'
        s += '      ' + then_str.replace('\n', '\n    ') + '\n'
        s += '  else:\n'
        s += '      ' + else_str.replace('\n', '\n    ') + '\n'
        s += '  )\n'
        val_dict[self] = s

    def diff_up(self, val_dict, der_dict):
        if_val = val_dict[self._if_arg]
        der_dict[self._if_arg] = 0
        if if_val:
            if self._then_arg.is_leaf():
                val = self._then_arg.value
                val_dict[self._then_arg] = val
                if self._then_arg not in der_dict:
                    der_dict[self._then_arg] = 0
            else:
                val = val_dict[self._then_arg]
                der_dict[self._then_arg] = 0
            if self._else_arg.is_leaf():
                _val = self._else_arg.value
                val_dict[self._else_arg] = _val
                if self._else_arg not in der_dict:
                    der_dict[self._else_arg] = 0
            else:
                der_dict[self._else_arg] = 0
        else:
            if self._else_arg.is_leaf():
                val = self._else_arg.value
                val_dict[self._else_arg] = val
                if self._else_arg not in der_dict:
                    der_dict[self._else_arg] = 0
            else:
                val = val_dict[self._else_arg]
                der_dict[self._else_arg] = 0
            if self._then_arg.is_leaf():
                _val = self._then_arg.value
                val_dict[self._then_arg] = _val
                if self._then_arg not in der_dict:
                    der_dict[self._then_arg] = 0
            else:
                der_dict[self._then_arg] = 0
        val_dict[self] = val

    def diff_up_symbolic(self, val_dict, der_dict):
        der_dict[self._if_arg] = 0
        if_val = val_dict[self._if_arg]
        if self._then_arg.is_leaf():
            then_val = self._then_arg
            val_dict[self._then_arg] = then_val
            if self._then_arg not in der_dict:
                der_dict[self._then_arg] = 0
        else:
            then_val = val_dict[self._then_arg]
            der_dict[self._then_arg] = 0
        if self._else_arg.is_leaf():
            else_val = self._else_arg
            val_dict[self._else_arg] = else_val
            if self._else_arg not in der_dict:
                der_dict[self._else_arg] = 0
        else:
            else_val = val_dict[self._else_arg]
            der_dict[self._else_arg] = 0
        val_dict[self] = if_else(if_val, then_val, else_val)

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._then_arg] += if_else(val_dict[self._if_arg], der, 0)
        der_dict[self._else_arg] += if_else(val_dict[self._if_arg], 0, der)

    def get_rpn(self, rpn_map, leaf_ndx_map):
        if self._if_arg.is_leaf():
            rpn_map[self] = _rpn = [leaf_ndx_map[self._if_arg]]
        else:
            rpn_map[self] = _rpn = rpn_map[self._if_arg]
        if self._then_arg.is_leaf():
            _rpn.append(leaf_ndx_map[self._then_arg])
        else:
            _rpn.extend(rpn_map[self._then_arg])
        if self._else_arg.is_leaf():
            _rpn.append(leaf_ndx_map[self._else_arg])
        else:
            _rpn.extend(rpn_map[self._else_arg])
        _rpn.append(OperationEnum.if_else.value)


class InequalityOperator(Operator):
    __slots__ = ('_lb', '_ub', '_body')

    def __init__(self, body, lb, ub):
        self._body = body
        self._lb = lb
        self._ub = ub

    def evaluate(self, val_dict):
        if self._body.is_leaf():
            body_val = self._body.value
        else:
            body_val = val_dict[self._body]
        val_dict[self] = (self._lb.value <= body_val <= self._ub.value)

    def operands(self):
        yield self._body
        yield self._lb
        yield self._ub

    def _str(self, val_dict):
        if self._body.is_leaf():
            body_val = self._body._str()
        else:
            body_val = val_dict[self._body]
        val_dict[self] = '(' + self._lb._str() + ' <= ' + body_val + ' <= ' + self._ub._str() + ')'

    def diff_up(self, val_dict, der_dict):
        if self._body.is_leaf():
            body_val = self._body.value
            val_dict[self._body] = body_val
            if self._body not in der_dict:
                der_dict[self._body] = 0
        else:
            body_val = val_dict[self._body]
            der_dict[self._body] = 0
        val_dict[self] = (self._lb.value <= body_val <= self._ub.value)

    def diff_up_symbolic(self, val_dict, der_dict):
        if self._body.is_leaf():
            body_val = self._body
            val_dict[self._body] = body_val
            if self._body not in der_dict:
                der_dict[self._body] = 0
        else:
            body_val = val_dict[self._body]
            der_dict[self._body] = 0
        val_dict[self] = inequality(body_val, self._lb.value, self._ub.value)

    def diff_down(self, val_dict, der_dict):
        pass

    def get_rpn(self, rpn_map, leaf_ndx_map):
        if self._body.is_leaf():
            rpn_map[self] = _rpn = [leaf_ndx_map[self._body]]
        else:
            rpn_map[self] = _rpn = rpn_map[self._body]
        _rpn.append(leaf_ndx_map[self._lb])
        _rpn.append(leaf_ndx_map[self._ub])
        _rpn.append(OperationEnum.inequality.value)


class SignOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'sign'
    operation_enum = OperationEnum.sign.value

    def diff_down(self, val_dict, der_dict):
        pass

    @staticmethod
    def operation(val):
        return sign(val)


class AbsOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'abs'
    operation_enum = OperationEnum.abs.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der * if_else(if_statement=inequality(body=val_dict[self._operand], lb=0),
                                                 then_statement=Float(1), else_statement=Float(-1))

    @staticmethod
    def operation(val):
        return abs(val)


class ExpOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'exp'
    operation_enum = OperationEnum.exp.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der * exp(val_dict[self._operand])

    @staticmethod
    def operation(val):
        return exp(val)


class LogOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'log'
    operation_enum = OperationEnum.log.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der / val_dict[self._operand]

    @staticmethod
    def operation(val):
        return log(val)


class SinOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'sin'
    operation_enum = OperationEnum.sin.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der * cos(val_dict[self._operand])

    @staticmethod
    def operation(val):
        return sin(val)


class CosOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'cos'
    operation_enum = OperationEnum.cos.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] -= der * sin(val_dict[self._operand])

    @staticmethod
    def operation(val):
        return cos(val)


class TanOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'tan'
    operation_enum = OperationEnum.tan.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der / (cos(val_dict[self._operand])**2)

    @staticmethod
    def operation(val):
        return tan(val)


class AsinOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'asin'
    operation_enum = OperationEnum.asin.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der / (1 - val_dict[self._operand]**2)**0.5

    @staticmethod
    def operation(val):
        return asin(val)


class AcosOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'acos'
    operation_enum = OperationEnum.acos.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] -= der / (1 - val_dict[self._operand]**2)**0.5

    @staticmethod
    def operation(val):
        return acos(val)


class AtanOperator(UnaryOperator):
    __slots__ = ()

    str_repn = 'atan'
    operation_enum = OperationEnum.atan.value

    def diff_down(self, val_dict, der_dict):
        der = der_dict[self]
        der_dict[self._operand] += der / (1 + val_dict[self._operand]**2)

    @staticmethod
    def operation(val):
        return atan(val)


def value(obj):
    if type(obj) in native_numeric_types:
        return obj
    return obj.evaluate()


def is_variable_type(obj):
    """
    Returns True if the object is a variable.

    Parameters
    ----------
    obj: ExpressionBase
        Also accepts floats and ints

    Returns
    -------
    bool
    """
    if type(obj) in native_numeric_types:
        return False
    return obj.is_variable_type()


class ConditionalExpression(object):
    def __init__(self):
        self._conditions = list()
        self._exprs = list()

    def add_condition(self, condition, expr):
        assert condition.is_relational()
        self._conditions.append(condition)
        self._exprs.append(expr)

    def add_final_expr(self, expr):
        self._conditions.append(Float(1))
        self._exprs.append(expr)

    def evaluate(self):
        for i, cond in enumerate(self._conditions):
            if cond.evaluate():
                return self._exprs[i].evaluate()

    def reverse_ad(self):
        for i, cond in enumerate(self._conditions):
            if cond.evaluate():
                return self._exprs[i].reverse_ad()

    def __str__(self):
        i = 0
        s = '\n'
        for _cond, _expr in zip(self._conditions, self._exprs):
            if i == 0:
                s += 'if '
            else:
                s += 'elif '
            s += str(_cond)
            s += ':\n    '
            s += str(_expr)
            s += '\n'
            i += 1
        return s

    def __repr__(self):
        return self.__str__()



# coding: utf-8

"""
The wntr.network.io module includes functions that convert the water network
model to other data formats, create a water network model from file, and write 
the water network model to a file.
"""
import logging
import json
import networkx as nx

import wntr.epanet
from wntr.epanet.util import FlowUnits
import wntr.network.model
from wntr.gis.network import WaterNetworkGIS
try:
    import geopandas as gpd
    has_geopandas = True
except ModuleNotFoundError:
    gpd = None
    has_geopandas = False
    
logger = logging.getLogger(__name__)


def to_dict(wn) -> dict:
    """
    Convert a WaterNetworkModel into a dictionary
    
    Parameters
    ----------
    wn : WaterNetworkModel
        Water network model

    Returns
    -------
    dict
        Dictionary representation of the WaterNetworkModel
        
    """
    from wntr import __version__

    controls = list()
    for k, c in wn._controls.items():
        cc = c.to_dict()
        if "name" in cc.keys() and not cc["name"]:
            cc["name"] = k
        controls.append(cc)
    d = dict(
        version="wntr-{}".format(__version__),
        comment="WaterNetworkModel - all values given in SI units",
        name=wn.name,
        references=wn._references.copy(),
        options=wn._options.to_dict(),
        curves=wn._curve_reg.to_list(),
        patterns=wn._pattern_reg.to_list(),
        nodes=wn._node_reg.to_list(),
        links=wn._link_reg.to_list(),
        sources=wn._sources.to_list(),
        controls=controls,
    )
    return d


def from_dict(d: dict, append=None):
    """
    Create or append a WaterNetworkModel from a dictionary

    Parameters
    ----------
    d : dict
        Dictionary representation of the water network model
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.

    Returns
    -------
    WaterNetworkModel
    
    """
    from wntr.epanet.io import _read_control_line, _EpanetRule

    keys = [
        "version",
        "comment",
        "name",
        "references",
        "options",
        "curves",
        "patterns",
        "nodes",
        "links",
        "sources",
        "controls",
    ]
    for key in keys:
        if key not in d.keys():
            logger.warning("Dictionary model missing key '{}'".format(key))
    if append is None:
        wn = wntr.network.model.WaterNetworkModel()
    else:
        wn = append
    if "name" in d:
        wn.name = d["name"]
    if "references" in d:
        wn._references = d["references"]
    if "options" in d:
        wn.options.__init__(**d["options"])
    if "curves" in d:
        for curve in d["curves"]:
            wn.add_curve(name=curve["name"], curve_type=curve["curve_type"], xy_tuples_list=curve["points"])
    if "patterns" in d:
        for pattern in d["patterns"]:
            wn.add_pattern(name=pattern["name"], pattern=pattern["multipliers"])
    if "nodes" in d:
        for node in d["nodes"]:
            name = node["name"]
            if node["node_type"] == "Junction":
                dl = node.setdefault("demand_timeseries_list")
                if dl is not None and len(dl) > 0:
                    base_demand = dl[0].setdefault("base_val", 0.0)
                    pattern_name = dl[0].setdefault("pattern_name")
                    demand_category = dl[0].setdefault("category")
                else:
                    base_demand = node.setdefault('base_demand',0.0)
                    pattern_name = node.setdefault('pattern_name')
                    demand_category = node.setdefault('demand_category')
                wn.add_junction(
                    name=name,
                    base_demand=base_demand,
                    demand_pattern=pattern_name,
                    elevation=node.setdefault("elevation", 0.0),
                    coordinates=node.setdefault("coordinates", list()),
                    demand_category=demand_category,
                )
                j = wn.get_node(name)
                j.emitter_coefficient = node.setdefault("emitter_coefficient")
                j.initial_quality = node.setdefault("initial_quality")
                j.minimum_pressure = node.setdefault("minimum_pressure")
                j.pressure_exponent = node.setdefault("pressure_exponent")
                j.required_pressure = node.setdefault("required_pressure")
                j.tag = node.setdefault("tag")
                
                j._leak = node.setdefault("leak", False)
                j._leak_area = node.setdefault("leak_area", 0.0)
                j._leak_discharge_coeff = node.setdefault("leak_discharge_coeff", 0.0)
                
                # custom additional attributes
                for attr in list(set(node.keys()) - set(dir(j))):
                    setattr( j, attr, node[attr] )
                if dl is not None and len(dl) > 1:
                    for i in range(1, len(dl)):
                        base_val = dl[i].setdefault("base_val", 0.0)
                        pattern_name = dl[i].setdefault("pattern_name")
                        category = dl[i].setdefault("category")
                        j.add_demand(base_val, pattern_name, category)
            elif node["node_type"] == "Tank":
                coordinates = node.setdefault("coordinates")

                wn.add_tank(
                    name,
                    elevation=node.setdefault("elevation"),
                    init_level=node.setdefault("init_level", node.setdefault("min_level", 0)),
                    min_level=node.setdefault("min_level", 0),
                    max_level=node.setdefault("max_level", node.setdefault("min_level", 0) + 10),
                    diameter=node.setdefault("diameter", 0),
                    min_vol=node.setdefault("min_vol", 0),
                    vol_curve=node.setdefault("vol_curve_name"),
                    overflow=node.setdefault("overflow", False),
                    coordinates=coordinates,
                )
                t = wn.get_node(name)
                t.initial_quality = node.setdefault("initial_quality", 0.0)
                if node.setdefault("mixing_fraction"):
                    t.mixing_fraction = node.setdefault("mixing_fraction")
                if node.setdefault("mixing_model"):
                    t.mixing_model = node.setdefault("mixing_model")
                t.bulk_coeff = node.setdefault("bulk_coeff")
                t.tag = node.setdefault("tag")
                # custom additional attributes
                for attr in list(set(node.keys()) - set(dir(t))):
                    setattr( t, attr, node[attr] )
            elif node["node_type"] == "Reservoir":
                wn.add_reservoir(
                    name,
                    base_head=node.setdefault("base_head"),
                    head_pattern=node.setdefault("head_pattern_name"),
                    coordinates=node.setdefault("coordinates"),
                )
                r = wn.get_node(name)
                r.initial_quality = node.setdefault("initial_quality", 0.0)
                r.tag = node.setdefault("tag")
                # custom additional attributes
                for attr in list(set(node.keys()) - set(dir(r))):
                    setattr( r, attr, node[attr] )
            else:
                raise ValueError("Illegal node type '{}'".format(node["node_type"]))
    if "links" in d:
        for link in d["links"]:
            name = link["name"]
            if link["link_type"] == "Pipe":
                wn.add_pipe(
                    name,
                    link["start_node_name"],
                    end_node_name=link["end_node_name"],
                    length=link.setdefault("length", 304.8),
                    diameter=link.setdefault("diameter", 0.3048),
                    roughness=link.setdefault("roughness", 100.0),
                    minor_loss=link.setdefault("minor_loss", 0.0),
                    initial_status=link.setdefault("initial_status", "OPEN"),
                    check_valve=link.setdefault("check_valve", False),
                )
                p = wn.get_link(name)
                p.bulk_coeff = link.setdefault("bulk_coeff")
                p.tag = link.setdefault("tag")
                p.vertices = link.setdefault("vertices", list())
                p.wall_coeff = link.setdefault("wall_coeff")
                # custom additional attributes
                for attr in list(set(link.keys()) - set(dir(p))):
                    setattr( p, attr, link[attr] )
            elif link["link_type"] == "Pump":
                pump_type = link.setdefault("pump_type", "POWER")
                wn.add_pump(
                    name,
                    link["start_node_name"],
                    link["end_node_name"],
                    pump_type=pump_type,
                    pump_parameter=link.setdefault("power")
                    if pump_type.lower() == "power"
                    else link.setdefault("pump_curve_name"),
                    speed=link.setdefault("base_speed", 1.0),
                    pattern=link.setdefault("speed_pattern_name"),
                    initial_status=link.setdefault("initial_status", "OPEN"),
                )
                p = wn.get_link(name)
                p.efficiency_curve_name = link.setdefault("efficiency_curve_name")
                p.energy_pattern = link.setdefault("energy_pattern")
                p.energy_price = link.setdefault("energy_price")
                p.initial_setting = link.setdefault("initial_setting")
                p.tag = link.setdefault("tag")
                p.vertices = link.setdefault("vertices", list())
                # custom additional attributes
                for attr in list(set(link.keys()) - set(dir(p))):
                    setattr( p, attr, link[attr] )
            elif link["link_type"] == "Valve":
                valve_type = link["valve_type"]
                wn.add_valve(
                    name,
                    link["start_node_name"],
                    link["end_node_name"],
                    diameter=link.setdefault("diameter", 0.3048),
                    valve_type=valve_type,
                    minor_loss=link.setdefault("minor_loss", 0),
                    initial_setting=link.setdefault("initial_setting", 0),
                    initial_status=link.setdefault("initial_status", "ACTIVE"),
                )
                v = wn.get_link(name)
                if valve_type.lower() == "gpv":
                    v.headloss_curve_name = link.setdefault("headloss_curve_name")
                v.vertices = link.setdefault("vertices", list())
                # custom additional attributes
                for attr in list(set(link.keys()) - set(dir(v))):
                    setattr( v, attr, link[attr] )
            else:
                raise ValueError("Illegal link type '{}'".format(link["link_type"]))
    if "sources" in d:
        for source in d["sources"]:
            wn.add_source(
                source["name"],
                node_name=source["node_name"],
                source_type=source["source_type"],
                quality=source["strength"],
                pattern=source["pattern"],
            )
    if "controls" in d:  # TODO: FIXME: FINISH
        control_count = 0
        for control in d["controls"]:
            ctrl_type = control["type"]
            if ctrl_type.lower() == "simple":
                control_count += 1
                control_name = "control " + str(control_count)
                ta = control["then_actions"][0].split()
                tstring = " ".join([ta[0], ta[1], ta[4]])
                cond = control["condition"].split()
                if cond[0].lower() == "system":
                    cstr = " ".join(["AT", cond[1], cond[3], cond[4] if len(cond) > 4 else ""])
                else:
                    cstr = " ".join(["IF", cond[0], cond[1], cond[3], cond[4]])
                ctrl = _read_control_line(tstring + " " + cstr, wn, FlowUnits.SI, control_name)
                wn.add_control(control_name, ctrl)
            elif ctrl_type.lower() == "rule":
                ctrllst = ["RULE"]
                control_name = control["name"]
                ctrllst.append(control["name"])
                ctrllst.append("IF")
                ctrllst.append(control["condition"])
                thenact = " AND ".join(control["then_actions"])
                ctrllst.append("THEN")
                ctrllst.append(thenact)
                if "else_actions" in control and control["else_actions"]:
                    ctrllst.append("ELSE")
                    ctrllst.append(" AND ".join(control["else_actions"]))
                ctrllst.append("PRIORITY")
                ctrllst.append(str(control["priority"]))
                ctrlstring = " ".join(ctrllst)
                c = _EpanetRule.parse_rules_lines([ctrlstring])
                wn.add_control(control_name, c[0].generate_control(wn))
            else:
                raise ValueError("Illegal control type '{}'".format(ctrl_type))
    return wn


def to_gis(wn, crs=None, pumps_as_points=False, valves_as_points=False):
    """
    Convert a WaterNetworkModel into GeoDataFrames
    
    Parameters
    ----------
    wn : WaterNetworkModel
        Water network model
    crs : str, optional
        Coordinate reference system, by default None
    pumps_as_points : bool, optional
        Represent pumps as points (True) or lines (False), by default False
    valves_as_points : bool, optional
        Represent valves as points (True) or lines (False), by default False
        
    Returns
    -------
    WaterNetworkGIS object that contains GeoDataFrames
        
    """
    gis_data = WaterNetworkGIS()
    gis_data._create_gis(wn, crs, pumps_as_points, valves_as_points)
    
    return gis_data


def from_gis(gis_data, append=None):
    """
    Create or append a WaterNetworkModel from GeoDataFrames
    
    Parameters
    ----------
    gis_data : WaterNetworkGIS or dictionary of GeoDataFrames
        GeoDataFrames containing water network attributes. If gis_data is a 
        dictionary, then the keys are junctions, tanks, reservoirs, pipes, 
        pumps, and valves. If the pumps or valves are Points, they will be 
        converted to Lines with the same start and end node location.
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.
        
    Returns
    -------
    WaterNetworkModel
        
    """
    if isinstance(gis_data, dict):
        gis_data = WaterNetworkGIS(gis_data)

    wn = gis_data._create_wn(append=append)
    
    return wn

def to_graph(wn, node_weight=None, link_weight=None, modify_direction=False):
    """
    Convert a WaterNetworkModel into a networkx MultiDiGraph
    
    Parameters
    ----------
    node_weight :  dict or pandas Series (optional)
        Node weights
    link_weight : dict or pandas Series (optional)
        Link weights.  
    modify_direction : bool (optional)
        If True, than if the link weight is negative, the link start and 
        end node are switched and the abs(weight) is assigned to the link
        (this is useful when weighting graphs by flowrate). If False, link 
        direction and weight are not changed.
        
    Returns
    --------
    networkx MultiDiGraph
    """
    G = nx.MultiDiGraph()

    for name, node in wn.nodes():
        G.add_node(name)
        nx.set_node_attributes(G, name="pos", values={name: node.coordinates})
        nx.set_node_attributes(G, name="type", values={name: node.node_type})

        if node_weight is not None:
            try:  # weight nodes
                value = node_weight[name]
                nx.set_node_attributes(G, name="weight", values={name: value})
            except:
                pass

    for name, link in wn.links():
        start_node = link.start_node_name
        end_node = link.end_node_name
        G.add_edge(start_node, end_node, key=name)
        nx.set_edge_attributes(G, name="type", values={(start_node, end_node, name): link.link_type})

        if link_weight is not None:
            try:  # weight links
                value = link_weight[name]
                if modify_direction and value < 0:  # change the direction of the link and value
                    G.remove_edge(start_node, end_node, name)
                    G.add_edge(end_node, start_node, name)
                    nx.set_edge_attributes(G, name="type", values={(end_node, start_node, name): link.link_type})
                    nx.set_edge_attributes(G, name="weight", values={(end_node, start_node, name): -value})
                else:
                    nx.set_edge_attributes(G, name="weight", values={(start_node, end_node, name): value})
            except:
                pass
    
    return G

def write_json(wn, path_or_buf, **kw_json,):
    """
    Write the WaterNetworkModel to a JSON file

    Parameters
    ----------
    path_or_buf : str or IO stream
        Name of the file or file pointer
    kw_json : keyword arguments
        Arguments to pass directly to `json.dump`
        
    """
    if isinstance(path_or_buf, str):
        with open(path_or_buf, "w") as fout:
            json.dump(to_dict(wn), fout, **kw_json)
    else:
        json.dump(to_dict(wn), path_or_buf, **kw_json)


def read_json(path_or_buf, append=None, **kw_json):
    """
    Create or append a WaterNetworkModel from a JSON file

    Parameters
    ----------
    f : str
        Name of the file or file pointer
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.
    kw_json : keyword arguments
        Keyword arguments to pass to `json.load`

    Returns
    -------
    WaterNetworkModel
    
    """
    if isinstance(path_or_buf, str):
        with open(path_or_buf, "r") as fin:
            d = json.load(fin, **kw_json)
    else:
        d = json.load(path_or_buf, **kw_json)
    return from_dict(d, append)


def write_inpfile(wn, filename: str, units=None, version: float = 2.2, 
                  force_coordinates: bool = False):
    """
    Write the WaterNetworkModel to an EPANET INP file

    .. note::

        By default, WNTR now uses EPANET version 2.2 for the EPANET simulator engine. Thus,
        The WaterNetworkModel will also write an EPANET 2.2 formatted INP file by default as well.
        Because the PDD analysis options will break EPANET 2.0, the ``version`` option will allow
        the user to force EPANET 2.0 compatibility at the expense of pressured-dependent analysis 
        options being turned off.

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        Water network model
    filename : string
        Name of the inp file.
    units : str, int or FlowUnits
        Name of the units being written to the inp file.
    version : float, {2.0, **2.2**}
        Optionally specify forcing EPANET 2.0 compatibility.
    force_coordinates : bool
        This only applies if `self.options.graphics.map_filename` is not `None`,
        and will force the COORDINATES section to be written even if a MAP file is
        provided. False by default, but coordinates **are** written by default since
        the MAP file is `None` by default.

    """
    if wn._inpfile is None:
        wn._inpfile = wntr.epanet.InpFile()
    if units is None:
        units = wn._options.hydraulic.inpfile_units
    wn._inpfile.write(filename, wn, units=units, version=version, force_coordinates=force_coordinates)


def read_inpfile(filename, append=None):
    """
    Create or append a WaterNetworkModel from an EPANET INP file

    Parameters
    ----------
    filename : string
        Name of the INP file.
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.
        
    Returns
    -------
    WaterNetworkModel
    
    """
    inpfile = wntr.epanet.InpFile()
    wn = inpfile.read(filename, wn=append)
    wn._inpfile = inpfile
    
    return wn


def write_geojson(wn, prefix: str, crs=None, pumps_as_points=True, 
                  valves_as_points=True):
    """
    Write the WaterNetworkModel to a set of GeoJSON files, one file for each
    network element.

    The GeoJSON only includes information from the water network model.
    To add results of a simulation or analysis, do:

    .. code::

        wn_gis = wn.to_gis()
        wn_gis.add_node_attributes(some_data_to_add, 'name_of_attribute')
        wn_gis.write_geojson(...)

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        Water network model
    prefix : str
        File prefix
    crs : str, optional
        Coordinate reference system, by default None
    pumps_as_points : bool, optional
        Represent pumps as points (True) or lines (False), by default False
    valves_as_points : bool, optional
        Represent valves as points (True) or lines (False), by default False
        
    """
    wn_gis = wn.to_gis(crs, pumps_as_points=pumps_as_points, 
                       valves_as_points=valves_as_points)
    wn_gis.write_geojson(prefix=prefix)


def read_geojson(files, index_col='name', append=None):
    """
    Create or append a WaterNetworkModel from GeoJSON files

    Parameters
    ----------
    files : dictionary
        Dictionary of GeoJSON filenames, where the keys are in the set 
        ('junction', 'tanks', 'reservoirs', 'pipes', 'pumps', 'valves') and 
        values are the corresponding GeoJSON filename
    index_col : str, optional
        Column that contains the element name
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.

    Returns
    -------
    WaterNetworkModel

    """
    gis_data = WaterNetworkGIS()
    gis_data.read_geojson(files, index_col=index_col)
    wn = gis_data._create_wn(append=append)

    return wn


def write_shapefile(wn, prefix: str, crs=None, pumps_as_points=True,
                    valves_as_points=True):
    """
    Write the WaterNetworkModel to a set of Esri Shapefiles, one directory for
    each network element.

    The Shapefiles only includes information from the water network model.
    To add results of a simulation or analysis, do:

    .. code::

        wn_gis = wn.to_gis()
        wn_gis.add_node_attributes(some_data_to_add, 'name_of_attribute')
        wn_gis.write_shapefile(...)

    Parameters
    ----------
    wn : wntr WaterNetworkModel
        Water network model
    prefix : str
        File and directory prefix
    crs : str, optional
        Coordinate reference system, by default None
    pumps_as_points : bool, optional
        Represent pumps as points (True) or lines (False), by default False
    valves_as_points : bool, optional
        Represent valves as points (True) or lines (False), by default False

    """
    wn_gis = wn.to_gis(crs, pumps_as_points=pumps_as_points,
                       valves_as_points=valves_as_points)
    wn_gis.write_shapefile(prefix=prefix)

def read_shapefile(files, index_col='name', append=None):
    """

    Create or append a WaterNetworkModel from Esri Shapefiles

    Parameters
    ----------
    files : dictionary
        Dictionary of Shapefile file or directory names, where the keys are in the set 
        ('junction', 'tanks', 'reservoirs', 'pipes', 'pumps', 'valves') and 
        values are the corresponding Shapefile filenames or directories
    index_col : str, optional
        Column that contains the element name
    append : WaterNetworkModel or None, optional
        Existing WaterNetworkModel to append.  If None, a new WaterNetworkModel 
        is created.

    Returns
    -------
    WaterNetworkModel

    """
    gis_data = WaterNetworkGIS()
    gis_data.read_shapefile(files,index_col=index_col)
    wn = gis_data._create_wn(append=append)

    return wn

def valid_gis_names(complete_list=True, truncate_names=None):
    """
    Valid column/field names for GeoJSON or Shapefiles
    
    Note that Shapefile field names are truncated to 10 characters 
    (set truncate_names=10)
    
    Parameters
    ----------
    complete_list : bool
        When true, returns both optional and required column/field names.
        When false, only returns required column/field names.
    truncate_names : None or int
        Truncate column/field names to specified number of characters, 
        set truncate=10 for Shapefiles.  None indicates no truncation.
        
    Returns
    ---------
    dict : Dictionary of valid GeoJSON or Shapefile column/field names
    """
    
    gis_data = WaterNetworkGIS()
    column_names = gis_data._valid_names(complete_list, truncate_names)

    return column_names



# coding: utf-8
"""
The wntr.msx.base module includes base classes for the multi-species water
quality model

Other than the enum classes, the classes in this module are all abstract
and/or mixin classes, and should not be instantiated directly.
"""

import logging
import os
from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
from enum import Enum
from typing import Any, Dict, Iterator, Generator
from wntr.epanet.util import NoteType

from wntr.utils.disjoint_mapping import DisjointMapping
from wntr.utils.enumtools import add_get

from numpy import (
    abs,
    arccos,
    arcsin,
    arctan,
    cos,
    cosh,
    exp,
    heaviside,
    log,
    log10,
    sign,
    sin,
    sinh,
    sqrt,
    tan,
    tanh,
)

def cot(x):
    return 1 / tan(x)

def arccot(x):
    return 1 / arctan(1 / x)

def coth(x):
    return 1 / tanh(x)

logger = logging.getLogger(__name__)

HYDRAULIC_VARIABLES = [
    {"name": "D", "note": "pipe diameter (feet or meters) "},
    {
        "name": "Kc",
        "note": "pipe roughness coefficient (unitless for Hazen-Williams or Chezy-Manning head loss formulas, millifeet or millimeters for Darcy-Weisbach head loss formula)",
    },
    {"name": "Q", "note": "pipe flow rate (flow units) "},
    {"name": "U", "note": "pipe flow velocity (ft/sec or m/sec) "},
    {"name": "Re", "note": "flow Reynolds number "},
    {"name": "Us", "note": "pipe shear velocity (ft/sec or m/sec) "},
    {"name": "Ff", "note": "Darcy-Weisbach friction factor "},
    {"name": "Av", "note": "Surface area per unit volume (area units/L) "},
    {"name": "Len", "note": "Pipe length (feet or meters)"},
]
"""Hydraulic variables defined in EPANET-MSX.
For reference, the valid values are provided in :numref:`table-msx-hyd-vars`.

.. _table-msx-hyd-vars:
.. table:: Valid hydraulic variables in multi-species quality model expressions.

    ============== ================================================
    **Name**       **Description**
    -------------- ------------------------------------------------
    ``D``          pipe diameter
    ``Kc``         pipe roughness coefficient
    ``Q``          pipe flow rate
    ``U``          pipe flow velocity
    ``Re``         flow Reynolds number
    ``Us``         pipe shear velocity
    ``Ff``         Darcy-Weisbach friction factor
    ``Av``         pipe surface area per unit volume
    ``Len``        pipe length
    ============== ================================================

:meta hide-value:
"""

EXPR_FUNCTIONS = dict(
    abs=abs,
    sgn=sign,
    sqrt=sqrt,
    step=heaviside,
    log=log,
    exp=exp,
    sin=sin,
    cos=cos,
    tan=tan,
    cot=cot,
    asin=arcsin,
    acos=arccos,
    atan=arctan,
    acot=arccot,
    sinh=sinh,
    cosh=cosh,
    tanh=tanh,
    coth=coth,
    log10=log10,
)
"""Mathematical functions available for use in expressions. See 
:numref:`table-msx-funcs` for a list and description of the 
different functions recognized. These names, case insensitive, are 
considered invalid when naming variables.

.. _table-msx-funcs:
.. table:: Functions defined for use in EPANET-MSX expressions.

    ============== ================================================================
    **Name**       **Description**
    -------------- ----------------------------------------------------------------
    ``abs``        absolute value
    ``sgn``        sign
    ``sqrt``       square-root
    ``step``       step function
    ``exp``        natural number, `e`, raised to a power
    ``log``        natural logarithm
    ``log10``      base-10 logarithm
    ``sin``        sine
    ``cos``        cosine
    ``tan``        tangent
    ``cot``        cotangent
    ``asin``       arcsine
    ``acos``       arccosine
    ``atan``       arctangent
    ``acot``       arccotangent
    ``sinh``       hyperbolic sine
    ``cosh``       hyperbolic cosine
    ``tanh``       hyperbolic tangent
    ``coth``       hyperbolic cotangent
    ``*``          multiplication
    ``/``          division
    ``+``          addition
    ``-``          negation and subtraction
    ``^``          power/exponents
    ``(``, ``)``   groupings and function parameters
    ============== ================================================================

:meta hide-value:
"""

RESERVED_NAMES = (
    tuple([v["name"] for v in HYDRAULIC_VARIABLES])
    + tuple([k.lower() for k in EXPR_FUNCTIONS.keys()])
    + tuple([k.upper() for k in EXPR_FUNCTIONS.keys()])
    + tuple([k.capitalize() for k in EXPR_FUNCTIONS.keys()])
)
"""WNTR reserved names. This includes the MSX hydraulic variables
(see :numref:`table-msx-hyd-vars`) and the MSX defined functions 
(see :numref:`table-msx-funcs`).

:meta hide-value:
"""

_global_dict = dict()
for k, v in EXPR_FUNCTIONS.items():
    _global_dict[k.lower()] = v
    _global_dict[k.capitalize()] = v
    _global_dict[k.upper()] = v
for v in HYDRAULIC_VARIABLES:
    _global_dict[v["name"]] = v["name"]


@add_get(abbrev=True)
class VariableType(Enum):
    """Type of reaction variable.

    The following types are defined, and aliases of just the first character
    are also defined.

    .. rubric:: Enum Members
    .. autosummary::

        SPECIES
        CONSTANT
        PARAMETER
        TERM
        RESERVED

    .. rubric:: Class Methods
    .. autosummary::
        :nosignatures:

        get

    """

    SPECIES = 3
    """Chemical or biological water quality species"""
    TERM = 4
    """Functional term, or named expression, for use in reaction expressions"""
    PARAMETER = 5
    """Reaction expression coefficient that is parameterized by tank or pipe"""
    CONSTANT = 6
    """Constant coefficient for use in reaction expressions"""
    RESERVED = 9
    """Variable that is either a hydraulic variable or other reserved word"""

    S = SPEC = SPECIES
    T = TERM
    P = PARAM = PARAMETER
    C = CONST = CONSTANT
    R = RES = RESERVED

    def __repr__(self):
        return repr(self.name)


@add_get(abbrev=True)
class SpeciesType(Enum):
    """Enumeration for species type

    .. warning:: These enum values are not the same as the MSX SpeciesType.

    .. rubric:: Enum Members

    .. autosummary::
        BULK
        WALL

    .. rubric:: Class Methods
    .. autosummary::
        :nosignatures:

        get

    """

    BULK = 1
    """Bulk species"""
    WALL = 2
    """Wall species"""

    B = BULK
    W = WALL

    def __repr__(self):
        return repr(self.name)
    

@add_get(abbrev=True)
class ReactionType(Enum):
    """Reaction type which specifies the location where the reaction occurs

    The following types are defined, and aliases of just the first character
    are also defined.

    .. rubric:: Enum Members
    .. autosummary::
        PIPE
        TANK

    .. rubric:: Class Methods
    .. autosummary::
        :nosignatures:

        get
    """

    PIPE = 1
    """Expression describes a reaction in pipes"""
    TANK = 2
    """Expression describes a reaction in tanks"""

    P = PIPE
    T = TANK

    def __repr__(self):
        return repr(self.name)
    

@add_get(abbrev=True)
class ExpressionType(Enum):
    """Type of reaction expression

    The following types are defined, and aliases of just the first character
    are also defined.

    .. rubric:: Enum Members
    .. autosummary::

        EQUIL
        RATE
        FORMULA

    .. rubric:: Class Methods
    .. autosummary::
        :nosignatures:

        get

    """

    EQUIL = 1
    """Equilibrium expressions where equation is being equated to zero"""
    RATE = 2
    """Rate expression where the equation expresses the rate of change of
    the given species with respect to time as a function of the other species
    in the model"""
    FORMULA = 3
    """Formula expression where the concentration of the named species is a
    simple function of the remaining species"""

    E = EQUIL
    R = RATE
    F = FORMULA

    def __repr__(self):
        return repr(self.name)


class ReactionBase(ABC):
    """Water quality reaction class.

    This is an abstract class for water quality reactions with partial concrete
    attribute and method definitions. All parameters and methods documented
    here must be defined by a subclass except for the following:

    .. rubric:: Concrete attributes

    The :meth:`__init__` method defines the following attributes concretely.
    Thus, a subclass should call :code:`super().__init__(species_name, note=note)`
    at the beginning of its own initialization.

    .. autosummary::

        ~ReactionBase._species_name
        ~ReactionBase.note

    .. rubric:: Concrete properties

    The species name is protected, and a reaction cannot be manually assigned
    a new species. Therefore, the following property is defined concretely.

    .. autosummary::

        species_name

    .. rubric:: Concrete methods

    The following methods are concretely defined, but can be overridden.

    .. autosummary::
        :nosignatures:

        __str__
        __repr__
    """

    def __init__(self, species_name: str, *, note: NoteType = None) -> None:
        """Reaction ABC init method.

        Make sure you call this method from your concrete subclass ``__init__`` method:
        
        .. code::
            
            super().__init__(species_name, note=note)

        Parameters
        ----------
        species_name : str
            Name of the chemical or biological species being modeled using this
            reaction
        note : (str | dict | ENcomment), optional keyword
            Supplementary information regarding this reaction, by default None
            (see-also :class:`~wntr.epanet.util.NoteType`)

        Raises
        ------
        TypeError
            If expression_type is invalid
        """
        if species_name is None:
            raise TypeError("The species_name cannot be None")
        self._species_name: str = str(species_name)
        """Protected name of the species"""
        self.note: NoteType = note
        """Optional note regarding the reaction (see :class:`~wntr.epanet.util.NoteType`)
        """

    @property
    def species_name(self) -> str:
        """Name of the species that has a reaction being defined."""
        return self._species_name

    @property
    @abstractmethod
    def reaction_type(self) -> Enum:
        """Reaction type (reaction location)."""
        raise NotImplementedError

    def __str__(self) -> str:
        """Return the name of the species and the reaction type, indicated by
        an arrow. E.g., 'HOCL->PIPE for chlorine reaction in pipes."""
        return "{}->{}".format(self.species_name, self.reaction_type.name)

    def __repr__(self) -> str:
        """Return a representation of the reaction from the dictionary
        representation - see :meth:`to_dict`"""
        return "{}(".format(self.__class__.__name__) + ", ".join(["{}={}".format(k, repr(getattr(self, k))) for k, v in self.to_dict().items()]) + ")"

    @abstractmethod
    def to_dict(self) -> dict:
        """Represent the object as a dictionary"""
        raise NotImplementedError


class VariableBase(ABC):
    """Multi-species water quality model variable

    This is an abstract class for water quality model variables with partial
    definition of concrete attributes and methods. Parameters and methods
    documented here must be defined by a subclass except for the following:

    .. rubric:: Concrete attributes

    The :meth:`__init__` method defines the following attributes concretely.
    Thus, a subclass should call :code:`super().__init__()` at the beginning
    of its own initialization.

    .. autosummary::

        ~VariableBase.name
        ~VariableBase.note

    .. rubric:: Concrete methods

    The following methods are concretely defined, but can be overridden.

    .. autosummary::
        :nosignatures:

        __str__
        __repr__
    """

    def __init__(self, name: str, *, note: NoteType = None) -> None:
        """Variable ABC init method.

        Make sure you call this method from your concrete subclass ``__init__`` method:
        
        .. code::
            
            super().__init__(name, note=note)

        Parameters
        ----------
        name : str
            Name/symbol for the variable. Must be a valid MSX variable name
        note : (str | dict | ENcomment), optional keyword
            Supplementary information regarding this variable, by default None
            (see-also :class:`~wntr.epanet.util.NoteType`)

        Raises
        ------
        KeyExistsError
            Name is already taken
        ValueError
            Name is a reserved word
        """
        if name in RESERVED_NAMES:
            raise ValueError("Name cannot be a reserved name")
        self.name: str = name
        """Name/ID of this variable, must be a valid EPANET/MSX ID"""
        self.note: NoteType = note
        """Optional note regarding the variable (see :class:`~wntr.epanet.util.NoteType`)
        """

    @property
    @abstractmethod
    def var_type(self) -> Enum:
        """Type of reaction variable"""
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        """Represent the object as a dictionary"""
        return dict(name=self.name)

    def __str__(self) -> str:
        """Return the name of the variable"""
        return self.name

    def __repr__(self) -> str:
        """Return a representation of the variable from the dictionary
        representation - see :meth:`to_dict`"""
        return "{}(".format(self.__class__.__name__) + ", ".join(["{}={}".format(k, repr(getattr(self, k))) for k, v in self.to_dict().items()]) + ")"


class ReactionSystemBase(ABC):
    """Abstract class for reaction systems, which contains variables and
    reaction expressions.

    This class contains the functions necessary to perform dictionary-style
    addressing of *variables* by their name. It does not allow dictionary-style
    addressing of reactions.

    This is an abstract class with some concrete attributes and methods.
    Parameters and methods documented here must be defined by a subclass
    except for the following:

    .. rubric:: Concrete attributes

    The :meth:`__init__` method defines the following attributes concretely.
    Thus, a subclass should call :code:`super().__init__()` or :code:`super().__init__(filename)`.

    .. autosummary::

        ~ReactionSystemBase._vars
        ~ReactionSystemBase._rxns

    .. rubric:: Concrete methods

    The following special methods are concretely provided to directly access
    items in the :attr:`_vars` attribute.

    .. autosummary::
        :nosignatures:

        __contains__
        __eq__
        __ne__
        __getitem__
        __iter__
        __len__
    """

    def __init__(self) -> None:
        """Constructor for the reaction system.
        
        Make sure you call this method from your concrete subclass ``__init__`` method:
        
        .. code::
            
            super().__init__()

        """
        self._vars: DisjointMapping = DisjointMapping()
        """Variables registry, which is mapped to dictionary functions on the
        reaction system object"""
        self._rxns: Dict[str, Any] = dict()
        """Reactions dictionary"""

    @abstractmethod
    def add_variable(self, obj: VariableBase) -> None:
        """Add a variable to the system"""
        raise NotImplementedError

    @abstractmethod
    def add_reaction(self, obj: ReactionBase) -> None:
        """Add a reaction to the system"""
        raise NotImplementedError

    @abstractmethod
    def variables(self) -> Generator[Any, Any, Any]:
        """Generator looping through all variables"""
        raise NotImplementedError

    @abstractmethod
    def reactions(self) -> Generator[Any, Any, Any]:
        """Generator looping through all reactions"""
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict:
        """Represent the reaction system as a dictionary"""
        raise NotImplementedError

    def __contains__(self, __key: object) -> bool:
        return self._vars.__contains__(__key)

    def __eq__(self, __value: object) -> bool:
        return self._vars.__eq__(__value)

    def __ne__(self, __value: object) -> bool:
        return self._vars.__ne__(__value)

    def __getitem__(self, __key: str) -> VariableBase:
        return self._vars.__getitem__(__key)

    def __iter__(self) -> Iterator:
        return self._vars.__iter__()

    def __len__(self) -> int:
        return self._vars.__len__()


class VariableValuesBase(ABC):
    """Abstract class for a variable's network-specific values

    This class should contain values for different pipes, tanks,
    etc., that correspond to a specific network for the reaction
    system. It can be used for initial concentration values, or
    for initial settings on parameters, but should be information
    that is clearly tied to a specific type of variable.

    This is a pure abstract class. All parameters
    and methods documented here must be defined by a subclass.
    """

    @property
    @abstractmethod
    def var_type(self) -> Enum:
        """Type of variable this object holds data for."""
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict:
        """Represent the object as a dictionary"""
        raise NotImplementedError


class NetworkDataBase(ABC):
    """Abstract class containing network specific data

    This class should be populated with things like initial quality,
    sources, parameterized values, etc.

    This is a pure abstract class. All parameters
    and methods documented here must be defined by a subclass.
    """

    @abstractmethod
    def to_dict(self) -> dict:
        """Represent the object as a dictionary"""
        raise NotImplementedError


class QualityModelBase(ABC):
    """Abstract multi-species water quality model

    This is an abstract class for a water quality model. All parameters and
    methods documented here must be defined by a subclass except for the
    following:

    .. rubric:: Concrete attributes

    The :meth:`__init__` method defines the following attributes concretely.
    Thus, a subclass should call :code:`super().__init__()` or :code:`super().__init__(filename)`.

    .. autosummary::

        ~QualityModelBase.name
        ~QualityModelBase.title
        ~QualityModelBase.description
        ~QualityModelBase._orig_file
        ~QualityModelBase._options
        ~QualityModelBase._rxn_system
        ~QualityModelBase._net_data
        ~QualityModelBase._wn
    """

    def __init__(self, filename=None):
        """QualityModel ABC init method.

        Make sure you call this method from your concrete subclass ``__init__`` method:
        
        .. code::
            
            super().__init__(filename=filename)

        Parameters
        ----------
        filename : str, optional
            File to use to populate the initial data
        """
        self.name: str = None if filename is None else os.path.splitext(os.path.split(filename)[1])[0]
        """Name for the model, or the MSX model filename (no spaces allowed)"""
        self.title: str = None
        """Title line from the MSX file, must be a single line"""
        self.description: str = None
        """Longer description; note that multi-line descriptions may not be 
        represented well in dictionary form"""
        self._orig_file: str = filename
        """Protected original filename, if provided in the constructor"""
        self._options = None
        """Protected options data object"""
        self._rxn_system: ReactionSystemBase = None
        """Protected reaction system object"""
        self._net_data: NetworkDataBase = None
        """Protected network data object"""
        self._wn = None
        """Protected water network object"""

    @property
    @abstractmethod
    def options(self):
        """Model options structure

        Concrete classes should implement this with the appropriate typing and
        also implement a setter method.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def reaction_system(self) -> ReactionSystemBase:
        """Reaction variables defined for this model

        Concrete classes should implement this with the appropriate typing.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def network_data(self) -> NetworkDataBase:
        """Network-specific values added to this model

        Concrete classes should implement this with the appropriate typing.
        """
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict:
        """Represent the object as a dictionary"""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_dict(self, data: dict) -> "QualityModelBase":
        """Create a new model from a dictionary

        Parameters
        ----------
        data : dict
            Dictionary representation of the model

        Returns
        -------
        QualityModelBase
            New concrete model
        """
        raise NotImplementedError



"""
The wntr.network.layer module includes methods to generate network layers
(information that is not stored in the water network model or the graph).
"""
import numpy as np
import pandas as pd
import random

def generate_valve_layer(wn, placement_type='strategic', n=1, seed=None):
    """
    Generate valve layer data, which can be used in valve segmentation analysis.

    Parameters
    -----------
    wn : wntr WaterNetworkModel
        A WaterNetworkModel object
        
    placement_type : string
        Options include 'strategic' and 'random'.  
        
        - If 'strategic', n is the number of pipes from each node that do not 
          contain a valve. In this case, n is generally 0, 1 or 2 
          (i.e. N, N-1, N-2 valve placement).
        - If 'random', then n randomly placed valves are used to define the 
          valve layer.
        
    n : int
        
        - If 'strategic', n is the number of pipes from each node that do not 
          contain a valve.
        - If 'random', n is the number of number of randomly placed valves.
        
    seed : int or None
        Random seed
       
    Returns
    ---------
    valve_layer : pandas DataFrame
        Valve layer, defined by node and link pairs (for example, valve 0 is 
        on link A and protects node B). The valve_layer DataFrame is indexed by
        valve number, with columns named 'node' and 'link'.
    """
    
    if seed is not None:
        np.random.seed(seed)
        random.seed(seed)
    valve_layer = []
    if placement_type=='random':
        all_valves = []
        for pipe_name, pipe in wn.pipes():
            all_valves.append((pipe_name, pipe.start_node_name))
            all_valves.append((pipe_name, pipe.end_node_name))

        for valve_tuple in random.sample(all_valves, n):
            pipe_name, node_name = valve_tuple
            valve_layer.append([pipe_name, node_name])
            
    elif placement_type == 'strategic':
        for node_name, node in wn.nodes():
            links = wn.get_links_for_node(node_name)
            for l in np.random.choice(links, max(len(links)-n,0), replace=False):
                valve_layer.append([l, node_name])
            
    valve_layer = pd.DataFrame(valve_layer, columns=['link', 'node'])  
    
    return valve_layer


"""
The wntr.epanet.toolkit module is a Python extension for the EPANET 
Programmers Toolkit DLLs.
"""
import ctypes
import logging
import os
import os.path
import platform
import sys
from ctypes import byref

if sys.version_info[0:2] <= (3, 11):
    from pkg_resources import resource_filename
else:
    from importlib.resources import files

from .exceptions import EN_ERROR_CODES, EpanetException
from .util import SizeLimits

logger = logging.getLogger(__name__)

epanet_toolkit = "wntr.epanet.toolkit"

if os.name in ["nt", "dos"]:
    libepanet = "libepanet/windows-x64/epanet22.dll"
elif sys.platform in ["darwin"]:
    if 'arm' in platform.platform().lower():
        libepanet = "libepanet/darwin-arm/libepanet2.dylib"
    else:
        libepanet = "libepanet/darwin-x64/libepanet22.dylib"
else:
    libepanet = "libepanet/linux-x64/libepanet22.so"


def ENgetwarning(code, sec=-1):
    if sec >= 0:
        hours = int(sec / 3600.0)
        sec -= hours * 3600
        mm = int(sec / 60.0)
        sec -= mm * 60
        header = "%3d:%.2d:%.2d" % (hours, mm, sec)
    else:
        header = "{}".format(code)
    if code < 100:
        msg = EN_ERROR_CODES.get(code, "Unknown warning %s")
    else:
        raise EpanetException(code)

    return msg % header


def runepanet(inpfile, rptfile=None, binfile=None):
    """Run an EPANET command-line simulation

    Parameters
    ----------
    inpfile : str
        The input file name
    """
    file_prefix, file_ext = os.path.splitext(inpfile)
    if rptfile is None:
        rptfile = file_prefix + ".rpt"
    if binfile is None:
        binfile = file_prefix + ".bin"

    enData = ENepanet()
    enData.ENopen(inpfile, rptfile, binfile)
    enData.ENsolveH()
    enData.ENsolveQ()
    try:
        enData.ENreport()
    except:
        pass
    enData.ENclose()


class ENepanet:
    """Wrapper class to load the EPANET DLL object, then perform operations on
    the EPANET object that is created when a file is loaded.

    This simulator is thread safe **only** for EPANET `version=2.2`.

    Parameters
    ----------
    inpfile : str
        Input file to use
    rptfile : str
        Output file to report to
    binfile : str
        Results file to generate
    version : float
        EPANET version to use (either 2.0 or 2.2)
    """

    def __init__(self, inpfile="", rptfile="", binfile="", version=2.2):
        self.ENlib = None
        self.errcode = 0
        self.errcodelist = []
        self.cur_time = 0

        self.Warnflag = False
        self.Errflag = False
        self.fileLoaded = False

        self.inpfile = inpfile
        self.rptfile = rptfile
        self.binfile = binfile

        try:
            if sys.version_info[0:2] <= (3, 11):
                if float(version) == 2.0:
                    libname = libepanet.replace('epanet22.','epanet20.')
                    if 'arm' in platform.platform():
                        raise NotImplementedError('ARM-based processors not supported for version 2.0 of EPANET. Please use version=2.2')
                else:
                    libname = libepanet
                libname = resource_filename(__name__, libname)
                if os.name in ["nt", "dos"]:
                    self.ENlib = ctypes.windll.LoadLibrary(libname)
                else:
                    self.ENlib = ctypes.cdll.LoadLibrary(libname)
            else:
                if float(version) == 2.0:
                    libname = libepanet.replace('epanet22.','epanet20.')
                    if 'arm' in platform.platform():
                        raise NotImplementedError('ARM-based processors not supported for version 2.0 of EPANET. Please use version=2.2')
                else:
                    libname = libepanet
                libname = files(__name__).joinpath(libname)
                if os.name in ["nt", "dos"]:
                    self.ENlib = ctypes.windll.LoadLibrary(libname)
                else:
                    self.ENlib = ctypes.cdll.LoadLibrary(libname)
        except:
            raise
        finally:
            if version >= 2.2:
                self._project = ctypes.c_uint64()
            else:
                self._project = None
        return

    def isOpen(self):
        """Checks to see if the file is open"""
        return self.fileLoaded

    def _error(self, *args):
        """Print the error text the corresponds to the error code returned"""
        if not self.errcode:
            return
        # errtxt = self.ENlib.ENgeterror(self.errcode)
        errtext = EN_ERROR_CODES.get(self.errcode, "unknown error")
        if "%" in errtext and len(args) == 1:
            errtext % args
        if self.errcode >= 100:
            self.Errflag = True
            logger.error("EPANET error {} - {}".format(self.errcode, errtext))
            raise EpanetException(self.errcode)
        else:
            self.Warnflag = True
            # warnings.warn(ENgetwarning(self.errcode))
            logger.warning("EPANET warning {} - {}".format(self.errcode, ENgetwarning(self.errcode, self.cur_time)))
            self.errcodelist.append(ENgetwarning(self.errcode, self.cur_time))
        return

    def ENopen(self, inpfile=None, rptfile=None, binfile=None):
        """
        Opens an EPANET input file and reads in network data

        Parameters
        ----------
        inpfile : str
            EPANET INP file (default to constructor value)
        rptfile : str
            Output file to create (default to constructor value)
        binfile : str
            Binary output file to create (default to constructor value)
        """
        if self._project is not None:
            if self.fileLoaded:
                self.EN_close(self._project)
            if self.fileLoaded:
                raise RuntimeError("File is loaded and cannot be closed")
            if inpfile is None:
                inpfile = self.inpfile
            if rptfile is None:
                rptfile = self.rptfile
            if binfile is None:
                binfile = self.binfile
            inpfile = inpfile.encode("latin-1")
            rptfile = rptfile.encode("latin-1")
            binfile = binfile.encode("latin-1")
            self.ENlib.EN_createproject(ctypes.byref(self._project))
            self.errcode = self.ENlib.EN_open(self._project, inpfile, rptfile, binfile)
            self._error()
            if self.errcode < 100:
                self.fileLoaded = True
            return
        else:
            if self.fileLoaded:
                self.ENclose()
            if self.fileLoaded:
                raise RuntimeError("File is loaded and cannot be closed")
            if inpfile is None:
                inpfile = self.inpfile
            if rptfile is None:
                rptfile = self.rptfile
            if binfile is None:
                binfile = self.binfile
            inpfile = inpfile.encode("latin-1")
            rptfile = rptfile.encode("latin-1")
            binfile = binfile.encode("latin-1")
            self.errcode = self.ENlib.ENopen(inpfile, rptfile, binfile)
            self._error()
            if self.errcode < 100:
                self.fileLoaded = True
            return

    def ENclose(self):
        """Frees all memory and files used by EPANET"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_close(self._project)
            self.ENlib.EN_deleteproject(self._project)
            self._project = None
            self._project = ctypes.c_uint64()
        else:
            self.errcode = self.ENlib.ENclose()
        self._error()
        if self.errcode < 100:
            self.fileLoaded = False
        return

    def ENsolveH(self):
        """Solves for network hydraulics in all time periods"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_solveH(self._project)
        else:
            self.errcode = self.ENlib.ENsolveH()
        self._error()
        return

    def ENsaveH(self):
        """Solves for network hydraulics in all time periods

        Must be called before ENreport() if no water quality simulation made.
        Should not be called if ENsolveQ() will be used.
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_saveH(self._project)
        else:
            self.errcode = self.ENlib.ENsaveH()
        self._error()
        return

    def ENopenH(self):
        """Sets up data structures for hydraulic analysis"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_openH(self._project)
        else:
            self.errcode = self.ENlib.ENopenH()
        self._error()
        return

    def ENinitH(self, iFlag):
        """Initializes hydraulic analysis

        Parameters
        -----------
        iFlag : 2-digit flag
            2-digit flag where 1st (left) digit indicates
            if link flows should be re-initialized (1) or
            not (0) and 2nd digit indicates if hydraulic
            results should be saved to file (1) or not (0)
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_initH(self._project, iFlag)
        else:
            self.errcode = self.ENlib.ENinitH(iFlag)
        self._error()
        return

    def ENrunH(self):
        """Solves hydraulics for conditions at time t

        This function is used in a loop with ENnextH() to run
        an extended period hydraulic simulation.
        See ENsolveH() for an example.

        Returns
        --------
        int
            Current simulation time (seconds)
        """
        lT = ctypes.c_long()
        if self._project is not None:
            self.errcode = self.ENlib.EN_runH(self._project, byref(lT))
        else:
            self.errcode = self.ENlib.ENrunH(byref(lT))
        self._error()
        self.cur_time = lT.value
        return lT.value

    def ENnextH(self):
        """Determines time until next hydraulic event

        This function is used in a loop with ENrunH() to run
        an extended period hydraulic simulation.
        See ENsolveH() for an example.

        Returns
        ---------
        int
            Time (seconds) until next hydraulic event (0 marks end of simulation period)
        """
        lTstep = ctypes.c_long()
        if self._project is not None:
            self.errcode = self.ENlib.EN_nextH(self._project, byref(lTstep))
        else:
            self.errcode = self.ENlib.ENnextH(byref(lTstep))
        self._error()
        return lTstep.value

    def ENcloseH(self):
        """Frees data allocated by hydraulics solver"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_closeH(self._project)
        else:
            self.errcode = self.ENlib.ENcloseH()
        self._error()
        return

    def ENsavehydfile(self, filename):
        """Copies binary hydraulics file to disk

        Parameters
        -------------
        filename : str
            Name of hydraulics file to output
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_savehydfile(self._project, filename.encode("latin-1"))
        else:
            self.errcode = self.ENlib.ENsavehydfile(filename.encode("latin-1"))
        self._error()
        return

    def ENusehydfile(self, filename):
        """Opens previously saved binary hydraulics file

        Parameters
        -------------
        filename : str
            Name of hydraulics file to use
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_usehydfile(self._project, filename.encode("latin-1"))
        else:
            self.errcode = self.ENlib.ENusehydfile(filename.encode("latin-1"))
        self._error()
        return

    def ENsolveQ(self):
        """Solves for network water quality in all time periods"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_solveQ(self._project)
        else:
            self.errcode = self.ENlib.ENsolveQ()
        self._error()
        return

    def ENopenQ(self):
        """Sets up data structures for water quality analysis"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_openQ(self._project)
        else:
            self.errcode = self.ENlib.ENopenQ()
        self._error()
        return

    def ENinitQ(self, iSaveflag):
        """Initializes water quality analysis

        Parameters
        -------------
        iSaveflag : int
             EN_SAVE (1) if results saved to file, EN_NOSAVE (0) if not
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_initQ(self._project, iSaveflag)
        else:
            self.errcode = self.ENlib.ENinitQ(iSaveflag)
        self._error()
        return

    def ENrunQ(self):
        """Retrieves hydraulic and water quality results at time t

        This function is used in a loop with ENnextQ() to run
        an extended period water quality simulation. See ENsolveQ() for
        an example.

        Returns
        -------
        int
            Current simulation time (seconds)
        """
        lT = ctypes.c_long()
        if self._project is not None:
            self.errcode = self.ENlib.EN_runQ(self._project, byref(lT))
        else:
            self.errcode = self.ENlib.ENrunQ(byref(lT))
        self._error()
        return lT.value

    def ENnextQ(self):
        """Advances water quality simulation to next hydraulic event

        This function is used in a loop with ENrunQ() to run
        an extended period water quality simulation. See ENsolveQ() for
        an example.

        Returns
        --------
        int
            Time (seconds) until next hydraulic event (0 marks end of simulation period)
        """
        lTstep = ctypes.c_long()
        if self._project is not None:
            self.errcode = self.ENlib.EN_nextQ(self._project, byref(lTstep))
        else:
            self.errcode = self.ENlib.ENnextQ(byref(lTstep))
        self._error()
        return lTstep.value

    def ENcloseQ(self):
        """Frees data allocated by water quality solver"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_closeQ(self._project)
        else:
            self.errcode = self.ENlib.ENcloseQ()
        self._error()
        return

    def ENreport(self):
        """Writes report to report file"""
        if self._project is not None:
            self.errcode = self.ENlib.EN_report(self._project)
        else:
            self.errcode = self.ENlib.ENreport()
        self._error()
        return

    def ENgetcount(self, iCode):
        """Retrieves the number of components of a given type in the network

        Parameters
        -------------
        iCode : int
            Component code (see toolkit.optComponentCounts)

        Returns
        ---------
        int
            Number of components in network

        """
        iCount = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getcount(self._project, iCode, byref(iCount))
        else:
            self.errcode = self.ENlib.ENgetcount(iCode, byref(iCount))
        self._error()
        return iCount.value

    def ENgetflowunits(self):
        """Retrieves flow units code

        Returns
        -----------
        Code of flow units in use (see toolkit.optFlowUnits)

        """
        iCode = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getflowunits(self._project, byref(iCode))
        else:
            self.errcode = self.ENlib.ENgetflowunits(byref(iCode))
        self._error()
        return iCode.value

    def ENgetnodeid(self, iIndex):
        """Gets the ID name of a node given its index.

        Parameters
        ----------
        iIndex : int
            a node's index (starting from 1).

        Returns
        -------
        str
            the node name
        """
        fValue = ctypes.create_string_buffer(SizeLimits.EN_MAX_ID.value)
        if self._project is not None:
            self.errcode = self.ENlib.EN_getnodeid(self._project, iIndex, byref(fValue))
        else:
            self.errcode = self.ENlib.ENgetnodeid(iIndex, byref(fValue))
        self._error()
        return str(fValue.value, "UTF-8")

    def ENgetnodeindex(self, sId):
        """Retrieves index of a node with specific ID

        Parameters
        -------------
        sId : int
            Node ID

        Returns
        ---------
        Index of node in list of nodes

        """
        iIndex = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getnodeindex(self._project, sId.encode("latin-1"), byref(iIndex))
        else:
            self.errcode = self.ENlib.ENgetnodeindex(sId.encode("latin-1"), byref(iIndex))
        self._error()
        return iIndex.value

    def ENgetnodetype(self, iIndex):
        """Retrieves a node's type given its index.

        Parameters
        ----------
        iIndex : int
            The index of the node

        Returns
        -------
        int
            the node type as an integer
        """
        fValue = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getnodetype(self._project, iIndex, byref(fValue))
        else:
            self.errcode = self.ENlib.ENgetnodetype(iIndex, byref(fValue))
        self._error()
        return fValue.value

    def ENgetnodevalue(self, iIndex, iCode):
        """Retrieves parameter value for a node

        Parameters
        -------------
        iIndex: int
            Node index
        iCode : int
            Node parameter code (see toolkit.optNodeParams)

        Returns
        ---------
        Value of node's parameter

        """
        fValue = ctypes.c_float()
        if self._project is not None:
            fValue = ctypes.c_double()
            self.errcode = self.ENlib.EN_getnodevalue(self._project, iIndex, iCode, byref(fValue))
        else:
            self.errcode = self.ENlib.ENgetnodevalue(iIndex, iCode, byref(fValue))
        self._error()
        return fValue.value

    def ENgetlinktype(self, iIndex):
        """Retrieves a link's type given its index.

        Parameters
        ----------
        iIndex : int
            The index of the link

        Returns
        -------
        int
            the link type as an integer
        """
        fValue = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getlinktype(self._project, iIndex, byref(fValue))
        else:
            self.errcode = self.ENlib.EN_getlinktype(iIndex, byref(fValue))
        self._error()
        return fValue.value

    def ENgetlinkindex(self, sId):
        """Retrieves index of a link with specific ID

        Parameters
        -------------
        sId : int
            Link ID

        Returns
        ---------
        Index of link in list of links

        """
        iIndex = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getlinkindex(self._project, sId.encode("latin-1"), byref(iIndex))
        else:
            self.errcode = self.ENlib.ENgetlinkindex(sId.encode("latin-1"), byref(iIndex))
        self._error()
        return iIndex.value

    def ENgetlinkvalue(self, iIndex, iCode):
        """Retrieves parameter value for a link

        Parameters
        -------------
        iIndex : int
            Link index
        iCode : int
            Link parameter code (see toolkit.optLinkParams)

        Returns
        ---------
        Value of link's parameter

        """
        fValue = ctypes.c_float()
        if self._project is not None:
            fValue = ctypes.c_double()
            self.errcode = self.ENlib.EN_getlinkvalue(self._project, iIndex, iCode, byref(fValue))
        else:
            self.errcode = self.ENlib.ENgetlinkvalue(iIndex, iCode, byref(fValue))
        self._error()
        return fValue.value

    def ENsetlinkvalue(self, iIndex, iCode, fValue):
        """Set the value on a link

        Parameters
        ----------
        iIndex : int
            the link index
        iCode : int
            the parameter enum integer
        fValue : float
            the value to set on the link
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_setlinkvalue(
                self._project, ctypes.c_int(iIndex), ctypes.c_int(iCode), ctypes.c_double(fValue)
            )
        else:
            self.errcode = self.ENlib.ENsetlinkvalue(ctypes.c_int(iIndex), ctypes.c_int(iCode), ctypes.c_float(fValue))
        self._error()

    def ENsetnodevalue(self, iIndex, iCode, fValue):
        """
        Set the value on a node

        Parameters
        ----------
        iIndex : int
            the node index
        iCode : int
            the parameter enum integer
        fValue : float
            the value to set on the node
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_setnodevalue(
                self._project, ctypes.c_int(iIndex), ctypes.c_int(iCode), ctypes.c_double(fValue)
            )
        else:
            self.errcode = self.ENlib.ENsetnodevalue(ctypes.c_int(iIndex), ctypes.c_int(iCode), ctypes.c_float(fValue))
        self._error()

    def ENsettimeparam(self, eParam, lValue):
        """Set a time parameter value

        Parameters
        ----------
        eParam : int
            the time parameter to set
        lValue : long
            the value to set, in seconds
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_settimeparam(self._project, ctypes.c_int(eParam), ctypes.c_long(lValue))
        else:
            self.errcode = self.ENlib.ENsettimeparam(ctypes.c_int(eParam), ctypes.c_long(lValue))
        self._error()

    def ENgettimeparam(self, eParam):
        """Get a time parameter value

        Parameters
        ----------
        eParam : int
            the time parameter to get

        Returns
        -------
        long
            the value of the time parameter, in seconds
        """
        lValue = ctypes.c_long()
        if self._project is not None:
            self.errcode = self.ENlib.EN_gettimeparam(self._project, ctypes.c_int(eParam), byref(lValue))
        else:
            self.errcode = self.ENlib.ENgettimeparam(ctypes.c_int(eParam), byref(lValue))
        self._error()
        return lValue.value

    def ENaddcontrol(self, iType: int, iLinkIndex: int, dSetting: float, iNodeIndex: int, dLevel: float) -> int:
        """Add a new simple control

        Parameters
        ----------
        iType : int
            the type of control
        iLinkIndex : int
            the index of the link
        dSetting : float
            the new link setting value
        iNodeIndex : int
            Set to 0 for time of day or timer
        dLevel : float
            the level to compare against

        Returns
        -------
        int
            the new control number
        """
        lValue = ctypes.c_int()
        if self._project is not None:
            self.errcode = self.ENlib.EN_addcontrol(
                self._project,
                ctypes.c_int(iType),
                ctypes.c_int(iLinkIndex),
                ctypes.c_double(dSetting),
                ctypes.c_int(iNodeIndex),
                ctypes.c_double(dLevel),
                byref(lValue),
            )
        else:
            self.errcode = self.ENlib.ENaddcontrol(
                ctypes.c_int(iType),
                ctypes.c_int(iLinkIndex),
                ctypes.c_double(dSetting),
                ctypes.c_int(iNodeIndex),
                ctypes.c_double(dLevel),
                byref(lValue),
            )
        self._error()
        return lValue.value

    def ENgetcontrol(self, iIndex: int):
        """Get values defined by a control.

        Parameters
        ----------
        iIndex : int
            the control number
        """
        iType = ctypes.c_int()
        iLinkIndex = ctypes.c_int()
        dSetting = ctypes.c_double()
        iNodeIndex = ctypes.c_int()
        dLevel = ctypes.c_double()
        if self._project is not None:
            self.errcode = self.ENlib.EN_getcontrol(
                self._project,
                ctypes.c_int(iIndex),
                byref(iType),
                byref(iLinkIndex),
                byref(dSetting),
                byref(iNodeIndex),
                byref(dLevel),
            )
        else:
            self.errcode = self.ENlib.ENgetcontrol(
                ctypes.c_int(iIndex), byref(iType), byref(iLinkIndex), byref(dSetting), byref(iNodeIndex), byref(dLevel)
            )
        self._error()
        return dict(
            index=iIndex,
            type=iType.value,
            linkindex=iLinkIndex.value,
            setting=dSetting.value,
            nodeindex=iNodeIndex.value,
            level=dLevel.value,
        )

    def ENsetcontrol(self, iIndex: int, iType: int, iLinkIndex: int, dSetting: float, iNodeIndex: int, dLevel: float):
        """Change values on a simple control

        Parameters
        ----------
        iIndex : int
            the control index
        iType : int
            the type of control comparison
        iLinkIndex : int
            the link being changed
        dSetting : float
            the setting to change to
        iNodeIndex : int
            the node being compared against, Set to 0 for time of day or timer
        dLevel : float
            the level being checked

        Warning
        -------
        There is an error in EPANET 2.2 that sets the `dLevel` parameter to 0.0 on Macs
        regardless of the value the user passes in. This means that to use this toolkit
        functionality on a Mac, the user must delete and create a new control to change
        the level.

        """
        if self._project is not None:
            try:
                self.errcode = self.ENlib.EN_setcontrol(
                    self._project,
                    ctypes.c_int(iIndex),
                    ctypes.c_int(iType),
                    ctypes.c_int(iLinkIndex),
                    ctypes.c_double(dSetting),
                    ctypes.c_int(iNodeIndex),
                    ctypes.c_double(dLevel),
                )
            except:
                self.errcode = self.ENlib.EN_setcontrol(
                    self._project,
                    ctypes.c_int(iIndex),
                    ctypes.c_int(iType),
                    ctypes.c_int(iLinkIndex),
                    ctypes.c_double(dSetting),
                    ctypes.c_int(iNodeIndex),
                    ctypes.c_float(dLevel),
                )
        else:
            self.errcode = self.ENlib.ENsetcontrol(
                ctypes.c_int(iIndex),
                ctypes.c_int(iType),
                ctypes.c_int(iLinkIndex),
                ctypes.c_double(dSetting),
                ctypes.c_int(iNodeIndex),
                ctypes.c_double(dLevel),
            )
        self._error()

    def ENdeletecontrol(self, iControlIndex):
        """Delete a control.

        Parameters
        ----------
        iControlIndex : int
            the simple control to delete
        """
        if self._project is not None:
            self.errcode = self.ENlib.EN_deletecontrol(self._project, ctypes.c_int(iControlIndex))
        else:
            self.errcode = self.ENlib.ENdeletecontrol(ctypes.c_int(iControlIndex))
        self._error()

    def ENsaveinpfile(self, inpfile):
        """Saves EPANET input file

        Parameters
        -------------
        inpfile : str
                    EPANET INP output file

        """

        inpfile = inpfile.encode("latin-1")
        if self._project is not None:
            self.errcode = self.ENlib.EN_saveinpfile(self._project, inpfile)
        else:
            self.errcode = self.ENlib.ENsaveinpfile(inpfile)
        self._error()

        return
